@@ -3,7 +3,7 @@
  *
  * Copyright (C) 1996 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
- * Copyright (C) 2004-2009 Wayne Davison
+ * Copyright (C) 2004-2013 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,14 +24,18 @@
  * are special. */
 
 #include "rsync.h"
+#include "ifuncs.h"
+#include "itypes.h"
 #include "io.h"
 
-extern int verbose;
 extern int am_root;
 extern int preserve_uid;
 extern int preserve_gid;
 extern int preserve_acls;
 extern int numeric_ids;
+extern gid_t our_gid;
+extern char *usermap;
+extern char *groupmap;
 
 #ifdef HAVE_GETGROUPS
 # ifndef GETGROUPS_T
@@ -39,24 +43,52 @@ extern int numeric_ids;
 # endif
 #endif
 
+#define NFLAGS_WILD_NAME_MATCH (1<<0)
+#define NFLAGS_NAME_MATCH (1<<1)
+
+union name_or_id {
+    const char *name;
+    id_t max_id;
+};
+
 struct idlist {
 	struct idlist *next;
-	const char *name;
+	union name_or_id u;
 	id_t id, id2;
 	uint16 flags;
 };
 
-static struct idlist *uidlist;
-static struct idlist *gidlist;
+static struct idlist *uidlist, *uidmap;
+static struct idlist *gidlist, *gidmap;
 
-static struct idlist *add_to_list(struct idlist **root, id_t id, const char *name,
+static id_t id_parse(const char *num_str)
+{
+	id_t tmp, num = 0;
+	const char *cp = num_str;
+
+	while (*cp) {
+		if (!isDigit(cp)) {
+		  invalid_num:
+			rprintf(FERROR, "Invalid ID number: %s\n", num_str);
+			exit_cleanup(RERR_SYNTAX);
+		}
+		tmp = num * 10 + *cp++ - '0';
+		if (tmp < num)
+			goto invalid_num;
+		num = tmp;
+	}
+
+	return num;
+}
+
+static struct idlist *add_to_list(struct idlist **root, id_t id, union name_or_id noiu,
 				  id_t id2, uint16 flags)
 {
 	struct idlist *node = new(struct idlist);
 	if (!node)
 		out_of_memory("add_to_list");
 	node->next = *root;
-	node->name = name;
+	node->u = noiu;
 	node->id = id;
 	node->id2 = id2;
 	node->flags = flags;
@@ -65,7 +97,7 @@ static struct idlist *add_to_list(struct idlist **root, id_t id, const char *nam
 }
 
 /* turn a uid into a user name */
-static const char *uid_to_name(uid_t uid)
+char *uid_to_user(uid_t uid)
 {
 	struct passwd *pass = getpwuid(uid);
 	if (pass)
@@ -74,7 +106,7 @@ static const char *uid_to_name(uid_t uid)
 }
 
 /* turn a gid into a group name */
-static const char *gid_to_name(gid_t gid)
+char *gid_to_group(gid_t gid)
 {
 	struct group *grp = getgrgid(gid);
 	if (grp)
@@ -82,20 +114,36 @@ static const char *gid_to_name(gid_t gid)
 	return NULL;
 }
 
-static uid_t map_uid(uid_t id, const char *name)
+/* Parse a user name or (optionally) a number into a uid */
+int user_to_uid(const char *name, uid_t *uid_p, BOOL num_ok)
 {
-	uid_t uid;
-	if (id != 0 && name_to_uid(name, &uid))
-		return uid;
-	return id;
+	struct passwd *pass;
+	if (!name || !*name)
+		return 0;
+	if (num_ok && name[strspn(name, "0123456789")] == '\0') {
+		*uid_p = id_parse(name);
+		return 1;
+	}
+	if (!(pass = getpwnam(name)))
+		return 0;
+	*uid_p = pass->pw_uid;
+	return 1;
 }
 
-static gid_t map_gid(gid_t id, const char *name)
+/* Parse a group name or (optionally) a number into a gid */
+int group_to_gid(const char *name, gid_t *gid_p, BOOL num_ok)
 {
-	gid_t gid;
-	if (id != 0 && name_to_gid(name, &gid))
-		return gid;
-	return id;
+	struct group *grp;
+	if (!name || !*name)
+		return 0;
+	if (num_ok && name[strspn(name, "0123456789")] == '\0') {
+		*gid_p = id_parse(name);
+		return 1;
+	}
+	if (!(grp = getgrnam(name)))
+		return 0;
+	*gid_p = grp->gr_gid;
+	return 1;
 }
 
 static int is_in_group(gid_t gid)
@@ -109,7 +157,6 @@ static int is_in_group(gid_t gid)
 	if (gid == last_in && last_out >= 0)
 		return last_out;
 	if (ngroups < -1) {
-		gid_t mygid = MY_GID();
 		if ((ngroups = getgroups(0, NULL)) < 0)
 			ngroups = 0;
 		gidset = new_array(GETGROUPS_T, ngroups+1);
@@ -119,12 +166,12 @@ static int is_in_group(gid_t gid)
 			ngroups = getgroups(ngroups, gidset);
 		/* The default gid might not be in the list on some systems. */
 		for (n = 0; n < ngroups; n++) {
-			if (gidset[n] == mygid)
+			if (gidset[n] == our_gid)
 				break;
 		}
 		if (n == ngroups)
-			gidset[ngroups++] = mygid;
-		if (verbose > 3) {
+			gidset[ngroups++] = our_gid;
+		if (DEBUG_GTE(OWN, 2)) {
 			int pos;
 			char *gidbuf = new_array(char, ngroups*21+32);
 			if (!gidbuf)
@@ -147,44 +194,59 @@ static int is_in_group(gid_t gid)
 	return last_out = 0;
 
 #else
-	static gid_t mygid = GID_NONE;
-	if (mygid == GID_NONE) {
-		mygid = MY_GID();
-		if (verbose > 3)
-			rprintf(FINFO, "process has gid %u\n", (unsigned)mygid);
-	}
-	return gid == mygid;
+	return gid == our_gid;
 #endif
 }
 
-/* Add a uid to the list of uids.  Only called on receiving side. */
-static struct idlist *recv_add_uid(uid_t id, const char *name)
+/* Add a uid/gid to its list of ids.  Only called on receiving side. */
+static struct idlist *recv_add_id(struct idlist **idlist_ptr, struct idlist *idmap,
+				  id_t id, const char *name)
 {
-	uid_t id2 = name ? map_uid(id, name) : id;
 	struct idlist *node;
+	union name_or_id noiu;
+	int flag;
+	id_t id2;
 
-	node = add_to_list(&uidlist, id, name, id2, 0);
+	noiu.name = name; /* ensure that add_to_list() gets the raw value. */
+	if (!name)
+		name = "";
 
-	if (verbose > 3) {
-		rprintf(FINFO, "uid %u(%s) maps to %u\n",
-			(unsigned)id, name ? name : "", (unsigned)id2);
+	for (node = idmap; node; node = node->next) {
+		if (node->flags & NFLAGS_WILD_NAME_MATCH) {
+			if (!wildmatch(node->u.name, name))
+				continue;
+		} else if (node->flags & NFLAGS_NAME_MATCH) {
+			if (strcmp(node->u.name, name) != 0)
+				continue;
+		} else if (node->u.max_id) {
+			if (id < node->id || id > node->u.max_id)
+				continue;
+		} else {
+			if (node->id != id)
+				continue;
+		}
+		break;
 	}
+	if (node)
+		id2 = node->id2;
+	else if (*name && id) {
+		if (idlist_ptr == &uidlist) {
+			uid_t uid;
+			id2 = user_to_uid(name, &uid, False) ? uid : id;
+		} else {
+			gid_t gid;
+			id2 = group_to_gid(name, &gid, False) ? gid : id;
+		}
+	} else
+		id2 = id;
 
-	return node;
-}
+	flag = idlist_ptr == &gidlist && !am_root && !is_in_group(id2) ? FLAG_SKIP_GROUP : 0;
+	node = add_to_list(idlist_ptr, id, noiu, id2, flag);
 
-/* Add a gid to the list of gids.  Only called on receiving side. */
-static struct idlist *recv_add_gid(gid_t id, const char *name)
-{
-	gid_t id2 = name ? map_gid(id, name) : id;
-	struct idlist *node;
-
-	node = add_to_list(&gidlist, id, name, id2,
-		!am_root && !is_in_group(id2) ? FLAG_SKIP_GROUP : 0);
-
-	if (verbose > 3) {
-		rprintf(FINFO, "gid %u(%s) maps to %u\n",
-			(unsigned)id, name ? name : "", (unsigned)id2);
+	if (DEBUG_GTE(OWN, 2)) {
+		rprintf(FINFO, "%sid %u(%s) maps to %u\n",
+			idlist_ptr == &uidlist ? "u" : "g",
+			(unsigned)id, name, (unsigned)id2);
 	}
 
 	return node;
@@ -193,23 +255,22 @@ static struct idlist *recv_add_gid(gid_t id, const char *name)
 /* this function is a definate candidate for a faster algorithm */
 uid_t match_uid(uid_t uid)
 {
-	static uid_t last_in, last_out;
+	static struct idlist *last = NULL;
 	struct idlist *list;
 
-	if (uid == 0)
-		return 0;
-
-	if (uid == last_in)
-		return last_out;
-
-	last_in = uid;
+	if (last && uid == last->id)
+		return last->id2;
 
 	for (list = uidlist; list; list = list->next) {
 		if (list->id == uid)
-			return last_out = list->id2;
+			break;
 	}
 
-	return last_out = uid;
+	if (!list)
+		list = recv_add_id(&uidlist, uidmap, uid, NULL);
+	last = list;
+
+	return list->id2;
 }
 
 gid_t match_gid(gid_t gid, uint16 *flags_ptr)
@@ -225,7 +286,7 @@ gid_t match_gid(gid_t gid, uint16 *flags_ptr)
 				break;
 		}
 		if (!list)
-			list = recv_add_gid(gid, NULL);
+			list = recv_add_id(&gidlist, gidmap, gid, NULL);
 		last = list;
 	}
 
@@ -239,6 +300,7 @@ const char *add_uid(uid_t uid)
 {
 	struct idlist *list;
 	struct idlist *node;
+	union name_or_id noiu;
 
 	if (uid == 0)	/* don't map root */
 		return NULL;
@@ -248,8 +310,9 @@ const char *add_uid(uid_t uid)
 			return NULL;
 	}
 
-	node = add_to_list(&uidlist, uid, uid_to_name(uid), 0, 0);
-	return node->name;
+	noiu.name = uid_to_user(uid);
+	node = add_to_list(&uidlist, uid, noiu, 0, 0);
+	return node->u.name;
 }
 
 /* Add a gid to the list of gids.  Only called on sending side. */
@@ -257,6 +320,7 @@ const char *add_gid(gid_t gid)
 {
 	struct idlist *list;
 	struct idlist *node;
+	union name_or_id noiu;
 
 	if (gid == 0)	/* don't map root */
 		return NULL;
@@ -266,8 +330,9 @@ const char *add_gid(gid_t gid)
 			return NULL;
 	}
 
-	node = add_to_list(&gidlist, gid, gid_to_name(gid), 0, 0);
-	return node->name;
+	noiu.name = gid_to_group(gid);
+	node = add_to_list(&gidlist, gid, noiu, 0, 0);
+	return node->u.name;
 }
 
 /* send a complete uid/gid mapping to the peer */
@@ -279,12 +344,12 @@ void send_id_list(int f)
 		int len;
 		/* we send sequences of uid/byte-length/name */
 		for (list = uidlist; list; list = list->next) {
-			if (!list->name)
+			if (!list->u.name)
 				continue;
-			len = strlen(list->name);
+			len = strlen(list->u.name);
 			write_varint30(f, list->id);
 			write_byte(f, len);
-			write_buf(f, list->name, len);
+			write_buf(f, list->u.name, len);
 		}
 
 		/* terminate the uid list with a 0 uid. We explicitly exclude
@@ -295,12 +360,12 @@ void send_id_list(int f)
 	if (preserve_gid || preserve_acls) {
 		int len;
 		for (list = gidlist; list; list = list->next) {
-			if (!list->name)
+			if (!list->u.name)
 				continue;
-			len = strlen(list->name);
+			len = strlen(list->u.name);
 			write_varint30(f, list->id);
 			write_byte(f, len);
-			write_buf(f, list->name, len);
+			write_buf(f, list->u.name, len);
 		}
 		write_varint30(f, 0);
 	}
@@ -318,7 +383,7 @@ uid_t recv_user_name(int f, uid_t uid)
 		free(name);
 		name = NULL;
 	}
-	node = recv_add_uid(uid, name); /* node keeps name's memory */
+	node = recv_add_id(&uidlist, uidmap, uid, name); /* node keeps name's memory */
 	return node->id2;
 }
 
@@ -334,7 +399,7 @@ gid_t recv_group_name(int f, gid_t gid, uint16 *flags_ptr)
 		free(name);
 		name = NULL;
 	}
-	node = recv_add_gid(gid, name); /* node keeps name's memory */
+	node = recv_add_id(&gidlist, gidmap, gid, name); /* node keeps name's memory */
 	if (flags_ptr && node->flags & FLAG_SKIP_GROUP)
 		*flags_ptr |= FLAG_SKIP_GROUP;
 	return node->id2;
@@ -361,17 +426,120 @@ void recv_id_list(int f, struct file_list *flist)
 
 	/* Now convert all the uids/gids from sender values to our values. */
 #ifdef SUPPORT_ACLS
-	if (preserve_acls && !numeric_ids)
+	if (preserve_acls && (!numeric_ids || usermap || groupmap))
 		match_acl_ids();
 #endif
-	if (am_root && preserve_uid && !numeric_ids) {
+	if (am_root && preserve_uid && (!numeric_ids || usermap)) {
 		for (i = 0; i < flist->used; i++)
 			F_OWNER(flist->files[i]) = match_uid(F_OWNER(flist->files[i]));
 	}
-	if (preserve_gid && (!am_root || !numeric_ids)) {
+	if (preserve_gid && (!am_root || !numeric_ids || groupmap)) {
 		for (i = 0; i < flist->used; i++) {
 			F_GROUP(flist->files[i]) = match_gid(F_GROUP(flist->files[i]),
 							     &flist->files[i]->flags);
 		}
 	}
 }
+
+void parse_name_map(char *map, BOOL usernames)
+{
+	struct idlist **idmap_ptr = usernames ? &uidmap : &gidmap;
+	struct idlist **idlist_ptr = usernames ? &uidlist : &gidlist;
+	char *colon, *cp = map + strlen(map);
+	union name_or_id noiu;
+	id_t id1;
+	uint16 flags;
+
+	/* Parse the list in reverse, so the order in the struct is right. */
+	while (1) {
+		while (cp > map && cp[-1] != ',') cp--;
+		if (!(colon = strchr(cp, ':'))) {
+			rprintf(FERROR, "No colon found in --%smap: %s\n",
+				usernames ? "user" : "group", cp);
+			exit_cleanup(RERR_SYNTAX);
+		}
+		if (!colon[1]) {
+			rprintf(FERROR, "No name found after colon --%smap: %s\n",
+				usernames ? "user" : "group", cp);
+			exit_cleanup(RERR_SYNTAX);
+		}
+		*colon = '\0';
+
+		if (isDigit(cp)) {
+			char *dash = strchr(cp, '-');
+			if (strspn(cp, "0123456789-") != (size_t)(colon - cp)
+			 || (dash && (!dash[1] || strchr(dash+1, '-')))) {
+				rprintf(FERROR, "Invalid number in --%smap: %s\n",
+					usernames ? "user" : "group", cp);
+				exit_cleanup(RERR_SYNTAX);
+			}
+			if (dash)
+				noiu.max_id = id_parse(dash+1);
+			else
+				noiu.max_id = 0;
+			flags = 0;
+			id1 = id_parse(cp);
+		} else if (strpbrk(cp, "*[?")) {
+			flags = NFLAGS_WILD_NAME_MATCH;
+			noiu.name = cp;
+			id1 = 0;
+		} else {
+			flags = NFLAGS_NAME_MATCH;
+			noiu.name = cp;
+			id1 = 0;
+		}
+
+		if (usernames) {
+			uid_t uid;
+			if (user_to_uid(colon+1, &uid, True))
+				add_to_list(idmap_ptr, id1, noiu, uid, flags);
+			else {
+				rprintf(FERROR,
+				    "Unknown --usermap name on receiver: %s\n",
+				    colon+1);
+			}
+		} else {
+			gid_t gid;
+			if (group_to_gid(colon+1, &gid, True))
+				add_to_list(idmap_ptr, id1, noiu, gid, flags);
+			else {
+				rprintf(FERROR,
+				    "Unknown --groupmap name on receiver: %s\n",
+				    colon+1);
+			}
+		}
+
+		if (cp == map)
+			break;
+
+		*--cp = '\0'; /* replace comma */
+	}
+
+	/* The 0 user/group doesn't get its name sent, so add it explicitly. */
+	recv_add_id(idlist_ptr, *idmap_ptr, 0,
+		    numeric_ids ? NULL : usernames ? uid_to_user(0) : gid_to_group(0));
+}
+
+#ifdef HAVE_GETGROUPLIST
+const char *getallgroups(uid_t uid, gid_t *gid_list, int *size_ptr)
+{
+	struct passwd *pw;
+	if ((pw = getpwuid(uid)) == NULL)
+		return "getpwuid failed";
+	/* Get all the process's groups, with the pw_gid group first. */
+	if (getgrouplist(pw->pw_name, pw->pw_gid, gid_list, size_ptr) < 0)
+		return "getgrouplist failed";
+	/* Paranoia: is the default group not first in the list? */
+	if (gid_list[0] != pw->pw_gid) {
+		int j;
+		for (j = 0; j < *size_ptr; j++) {
+			if (gid_list[j] == pw->pw_gid) {
+				gid_list[j] = gid_list[0];
+				gid_list[0] = pw->pw_gid;
+				break;
+			}
+		}
+	}
+	return NULL;
+}
+#endif
